@@ -6,17 +6,7 @@ macro_rules! emit {
     }};
 }
 
-macro_rules! punct {
-    ($g:expr,$span:expr,$n:expr) => {{
-        $g.wr.write_punct(Some($span), $n)?;
-    }};
-
-    ($g:expr,$n:expr) => {{
-        $g.wr.write_punct(None, $n)?;
-    }};
-}
-
-macro_rules! keyword {
+macro_rules! write {
     ($g:expr,$span:expr,$n:expr) => {{
         $g.wr.write_raw(Some($span), $n)?;
     }};
@@ -26,14 +16,6 @@ macro_rules! keyword {
     }};
 }
 
-macro_rules! formatting_space {
-    ($g:expr) => {{
-        if !$g.config.minify {
-            $g.wr.write_space()?;
-        }
-    }};
-}
-
 macro_rules! formatting_newline {
     ($g:expr) => {{
         if !$g.config.minify {
@@ -48,16 +30,40 @@ macro_rules! space {
     }};
 }
 
+macro_rules! formatting_space {
+    ($g:expr) => {{
+        if !$g.config.minify {
+            $g.wr.write_space()?;
+        }
+    }};
+}
+
 macro_rules! semi {
     ($g:expr) => {{
-        punct!($g, ";");
+        write!($g, ";");
     }};
 }
 
 macro_rules! formatting_semi {
     ($g:expr) => {{
         if !$g.config.minify {
-            punct!($g, ";");
+            write!($g, ";");
+        }
+    }};
+}
+
+macro_rules! increase_indent {
+    ($g:expr) => {{
+        if !$g.config.minify {
+            $g.wr.increase_indent();
+        }
+    }};
+}
+
+macro_rules! decrease_indent {
+    ($g:expr) => {{
+        if !$g.config.minify {
+            $g.wr.decrease_indent();
         }
     }};
 }
